from __future__ import print_function

import os.path
import shutil

from datetime import datetime
from subprocess import call
from util import render

from util import tempdir
from s3 import S3


class LambdaBuilder(object):
    """
    Create and upload ZIP archive to S3 for AWS lambda function from CFN.
    """

    def __init__(self, *zip_args, **zip_kwargs):
        self._zip_args = zip_args
        self._zip_kwargs = zip_kwargs
        self._created = {}  # "filename": "file body"
        self._copied = {}  # "filename": "source filename"

    def create_file(self, file_body, path, render_mapping=None):
        print("  * created `%s` from LambdaBuilder" % path)
        self._created[path] = file_body if render_mapping is None else render(file_body, render_mapping)
        return self

    def append(self, file_name, path=None, render_mapping=None):
        assert os.path.exists(file_name)
        assert os.path.isfile(file_name) or not render_mapping
        path = path or os.path.basename(file_name)

        if render_mapping:
            with open(file_name) as fp:
                self.create_file(fp.read(), path, render_mapping)
        else:
            print("  * copied `%s` from LambdaBuilder" % path)
            self._copied[path] = file_name

        return self

    def upload_to_s3(self, bucket_name, prefix=None, key_name=None):
        """
        Upload composed archive (.zip) to S3

        :param str prefix: without leading and trailing slashes
        """
        s3 = S3()

        prefix = prefix or "lambda-src"
        key_name = key_name or datetime.utcnow().strftime("%Y-%m-%dT%H:%M:%SZ.zip")

        uploaded_path = "%s/%s" % (prefix, key_name)

        with tempdir() as td:
            archive_name = self._compile(td.path, key_name)
            s3.upload(archive_name, bucket_name, uploaded_path)

        return uploaded_path

    def _compile(self, temp_path, archive_name):
        """
        Compiles to an zip archive and return its fullpath.
        Uses temp_path as working directory.

        :param temp_path:
        :param archive_name:
        :return: Fullpath to archive
        """
        old_umask = os.umask(0o022)

        build_path = os.path.join(temp_path, "build")
        os.mkdir(build_path, 0o755)

        archive_name = os.path.join(temp_path, archive_name)
        try:
            for filename, body in self._created.items():
                filename = os.path.join(build_path, filename)

                with open(filename, "wb") as fp:
                    fp.write(body)

            for filename, source in self._copied.items():
                filename = os.path.join(build_path, filename)

                if os.path.isdir(source):
                    shutil.copytree(source, filename)
                else:
                    shutil.copy(source, filename)

            call(["chmod", "-R", "go+rX", temp_path])
            call(["zip", "-qr", archive_name, ".", "-x", ".*", "*/.*", "__pycache__", "*/__pycache__"], cwd=build_path)

        finally:
            os.umask(old_umask)

        return archive_name



"""
Helper functions and wrapper classes to AWS CloudFormation.
"""
from __future__ import print_function

import boto3
import botocore.exceptions
import copy
import json
import os.path
import sys
import time

from aws import Aws
from s3 import S3
from exceptions import StackExists


class CfnStack(Aws):
    """
    Avatar to AWS CloudFormation, relies on its own region_name instead of environment.
    """
    QUERY_DELAY_SEC = 2

    def __init__(self, stack_name, region_name=None):
        """
        :param str stack_name: Name to the CFN Stack
        :param cfn_r: CFN Resource or None
        """
        region_name = region_name or self.get_region()
        
        self._cfn = boto3.resource('cloudformation', region_name=region_name)
        self._stack_name = stack_name
        self._stack = self._cfn.Stack(stack_name)

        self._capabilities = set()
        self._parameters = {}  # key: {cond: value}, hydrated to {"Key": key, cond: value}
        self._tags = {}

    def delete(self, cleanup_s3=False):
        """
        Delete the stack no matter it exists or not.
        """

        try:
            if cleanup_s3:
                self._stack.load()

                s3 = S3()
                for res in self._stack.resource_summaries.iterator():
                    if res.resource_type != "AWS::S3::Bucket":
                        continue

                    s3.cleanup_bucket(res.physical_resource_id)
        except botocore.exceptions.ClientError:
            pass

        self._stack.delete()

    def aws_capabilities(self):
        """
        Gets `capabilities`, valid values = {"CAPABILITY_IAM"}.
        @see http://docs.aws.amazon.com/AWSCloudFormation/latest/APIReference/API_CreateStack.html

        :return: list of capabilities
        """
        return list(self._capabilities)

    def with_capability(self, capability):
        """
        Enable capability

        :param str capability: capability
        :return: self
        :rtype: CfnStack
        """
        self._capabilities.add(str(capability))
        return self

    def without_capability(self, capability):
        """
        Disable capability, proceed silently if capability doesn't exist.
        Note: stack update may fail if required capability revoked in later updates.

        :param str capability: capability
        :return: self
        :rtype: CfnStack
        """
        try:
            self._capabilities.remove(str(capability))
        except KeyError:
            pass

        return self

    def aws_parameters(self):
        """
        Gets parameters in AWS Cloudformation format::

            [
                {
                    "ParameterKey": "somekey",
                    "ParameterValue": "somevalue"
                },
                {
                    "ParameterKey": "anotherkey",
                    "UsePreviousValue": True
                },
                ...
            ]

        :return: Parameters
        :rtype: list(dict)
        """
        def merge_dict(dict1, dict2):
            return dict(dict1.items() + dict2.items())

        return [
            merge_dict({"ParameterKey": k}, v) for k, v in self._parameters.items()
        ]

    def with_parameter(self, key, value=None):
        """
        Append parameter `key` to the list. Value would be `value` if it's not None, or `UsePreviousValue` otherwise.
        CfnStack guarantees Parameter Key uniqueness by storing them as dict{key: {cond: value}} internally.
        Therefore, updating a known parameter overrides it.

        :param key: Parameter key
        :param None|str value: Parameter value if not None, or trigger UsePreviousValue otherwise
        :return: self
        :rtype: CfnStack
        """
        key = str(key)

        if value is None:
            self._parameters[key] = {"UsePreviousValue": True}
        else:
            self._parameters[key] = {"ParameterValue": str(value)}

        return self

    def without_parameter(self, key):
        """
        Disable parameter, returns silently if key doesn't exist.

        :param str key: Parameter key
        :return: self
        :rtype: CfnStack
        """
        try:
            del self._parameters[str(key)]
        except KeyError:
            pass

        return self

    def carry_over_parameters(self):
        """
        Updates parameter list, so that all of them UsePreviousValue.
        Called after updating stack successfully.
        """
        for k in self._parameters:
            self.with_parameter(k, None)

    def aws_tags(self):
        """
        Gets tags in AWS Cloudformation format::

            [
                {"Key": key, "Value": value},
                ...
            ]

        :return: tags
        :rtype: list(dict)
        """
        return [{"Key": k, "Value": v} for k, v in self._tags.items()]

    def with_tag(self, key, value):
        """
        Append tag

        :param str key: Tag key
        :param str value: Tag value
        :return: self
        :rtype: CfnStack
        """
        self._tags[str(key)] = str(value)
        return self

    def without_tag(self, key):
        """
        Remove tag

        :param str key: Tag key
        :return: self
        :rtype: CfnStack
        """
        try:
            del self._tags[str(key)]
        except KeyError:
            pass

        return self

    def create(self, template, **kwargs):
        """
        @see cloudformation.create_stack()

        :param template:
        :param kwargs:
        :return: cloudformation stack
        """
        try:
            self._stack = self._cfn.create_stack(
                StackName=self._stack_name,
                TemplateBody=str(template),
                Parameters=self.aws_parameters(),
                Capabilities=self.aws_capabilities(),
                Tags=self.aws_tags(),
                **kwargs
            )
            self.carry_over_parameters()

        except botocore.exceptions.ClientError as e:
            if e.response.get('Error', {}).get('Code', '') == "AlreadyExistsException":
                raise StackExists()
            else:
                raise e

        return self

    def update(self, template=None, **kwargs):
        """
        @see cloudformation.update_stack()
        Calls CfnStack.carry_over_parameters if update() succeeds.

        :param template:
        :param kwargs:
        :return: cloudformation stack
        """
        params = {
            "Parameters": self.aws_parameters(),
            "Capabilities": self.aws_capabilities(),
            "Tags": self.aws_tags(),
        }

        if template is None:
            params.update({"UsePreviousTemplate": True})
        else:
            params.update({"TemplateBody": str(template)})

        params.update(kwargs)

        self._stack.update(**params)
        self.carry_over_parameters()

        return self

    def wait(self):
        """
        Waits and blocks until stack is in resolved state (*_COMPLETE).
        """
        if not self._stack:
            raise Exception("No stack available")

        stack = self._stack
        state_ready = (
                    'CREATE_COMPLETE',
                    'UPDATE_COMPLETE',
                    'UPDATE_ROLLBACK_COMPLETE')

        state_in_progress = (
                    'UPDATE_IN_PROGRESS',
                    'CREATE_IN_PROGRESS',
                    'UPDATE_COMPLETE_CLEANUP_IN_PROGRESS')

        state_fail = (
            "UPDATE_ROLLBACK_IN_PROGRESS",
        )

        while True:
            # The infinite loop, return on success, throw exception on error, pass to wait
            time.sleep(self.QUERY_DELAY_SEC)
            stack.load()
            if stack.stack_status in state_ready:
                # Completed
                break
            elif stack.stack_status in state_in_progress:
                pass
            elif stack.stack_status in state_fail:
                print("Stack operation failed (%s), try again later" % stack.stack_status)
                sys.exit(1)
            else:
                raise Exception("Unknown state: %s" % stack.stack_status)

        return self

    @property
    def outputs(self):
        return {i['OutputKey']: i['OutputValue'] for i in self._stack.outputs}

    @staticmethod
    def list(region_name=None):
        """
        Generator to list all cloudformation stacks

        :return: generator of Stacks
        :rtype: list(Stack)
        """
        cfn = boto3.resource('cloudformation', region_name=region_name)
        for stack in cfn.stacks.all():
            yield stack

    @classmethod
    def wait_for_stack(cls, stack_name):
        """
        Waits and blocks until stack is in resolved state (*_COMPLETE).
        """
        stack = CfnStack(stack_name)
        stack.wait()


class CfnTemplate(object):
    """
    Handles Cloud Formation Templates
    """
    KEYS_IGNORE = ('AWSTemplateFormatVersion',)
    KEYS_OVERWRITE = ('Description',)
    KEYS_KEY_MERGE = ('Parameters', 'Mappings', 'Conditions', 'Resources', 'Outputs')

    def load_template(self, f, deep_copy_dict=False):
        """
        Load template from f

        :param f: json string, plain file (json), and native data structure
        :param deep_copy_dict:
        :return: template as dict / list
        """
        if isinstance(f, str):
            try:
                return json.loads(f)
            except ValueError:
                pass

            # @TODO: Secure f ?
            f = os.path.join(self._template_path, f)
            if not f.endswith(".json"):
                f += ".json"

            with open(f) as fp:
                return json.load(fp)
        elif isinstance(f, file):
            return json.load(f)
        elif deep_copy_dict:
            return copy.deepcopy(f)

        return f

    def compose_template(self, base, *extends):
        """
        Loads `base` and `extend` respectively and combine them based on specifications
        on AWS CloudFormation templates. Deep copy on `base` if it's a dict.

        Doesn't rewind when base / extend is file.

        :param str|file|dict base: Base object or filename
        :param list[str|file|dict] extends: Extend object or filename

        :return: Combined object
        :rtype: dict
        """

        output = self.load_template(base, deep_copy_dict=True)

        for extend in extends:
            self.extend_template(output, extend)

        return output

    def extend_template(self, base, extend):
        """
        Extend properties from extend to base and return.
        Note: base is changed in the process.
        """
        base, extend = self.load_template(base), self.load_template(extend)

        if not base:
            base.update(extend)
            return base

        for key, value in extend.items():
            if key in self.KEYS_IGNORE:
                pass
            elif key in self.KEYS_OVERWRITE:
                base[key] = value
            elif key in self.KEYS_KEY_MERGE:
                for _k, _v in value.items():
                    base.setdefault(key, {})[_k] = _v
            else:
                print("Unknown key `%s` found.")

        return base

    def __init__(self, template_path, mapping=None):
        """
        :param template_path: Where to lookup templates
        :param mapping: Mapping table to render table
        """
        self._template_path = template_path
        self._template_dict = {}
        self._mapping = mapping or {}

    def with_documents(self, *templates):
        """
        Append templates to itself.
        @see CfnTemplate.load_template for supported formats.

        :param templates: templates
        :return: self
        :rtype: CfnTemplate
        """
        for template in templates:
            self.extend_template(self._template_dict, self.load_template(template))

        return self

    def with_mapping_key(self, key, value):
        """
        Append key=>value to mapping table, update if key already exists.

        :param key:
        :param value:
        :return: self
        :rtype: CfnTemplate
        """
        self._mapping[key] = value
        return self

    def without_mapping_key(self, key):
        """
        Remove key from mapping table, return silently if key doesn't exist.

        :param key:
        :return: self
        :rtype: CfnTemplate
        """
        try:
            del self._mapping[key]
        except KeyError:
            pass
        return self

    def with_updated_mapping(self, mapping):
        """
        Update multiple k-v mappings at once

        :param mapping:
        :return: self
        :rtype: CfnTemplate
        """
        self._mapping.update(mapping)
        return self

    def patch_resources(self, **kwargs):
        res_name, res_type, property, new_value = (kwargs.get(n, False) for n in ('Name', 'Type', 'Property', 'NewValue'))

        try:
            assert bool(res_name) ^ bool(res_type)
        except AssertionError:
            raise TypeError("keyword argment `Name` and `Type` are mutually exclusive.")

        try:
            assert property and new_value
        except AssertionError:
            raise TypeError("keyword argument `Property` and `NewValue` required.")

        if res_name:
            if res_name not in self._template_dict['Resources']:
                raise KeyError('resource `%s` not found' % res_name)

            self._template_dict['Resources'][res_name]['Properties'][property] = new_value
            return self

        for res_value in self._template_dict.get('Resources', {}).values():
            if res_value['Type'] == res_type:
                res_value['Properties'][property] = new_value

        return self

    def as_json(self, indent=2):
        """
        Return merged template in json without mapping

        :param indent:
        :return: template as json
        :rtype: str
        """
        from util import render

        return render(
            template=json.dumps(self._template_dict, indent=indent),
            mapping=self._mapping
        )

    def __str__(self):
        return self.as_json()



import boto3
import json
import urllib2

class Aws(object):
    """
    """
    TIMEOUT_MAGIC_URL = 1

    @classmethod
    def get_region_magic_url(cls):
        url = "http://169.254.169.254/latest/dynamic/instance-identity/document"
        try:
            return json.loads(urllib2.urlopen(url, timeout=cls.TIMEOUT_MAGIC_URL).read()).get('region')
        except (urllib2.HTTPError, urllib2.URLError, ValueError):
            pass

        return None

    @classmethod
    def get_region(cls):
        return boto3.Session().region_name or \
            cls.get_region_magic_url()

    @classmethod
    def get_region_shortname(cls, region_name=None):
        alternatives = {
            'use1': {'us-east-1', },
            'usw1': {'us-west-1', },
            'usw2': {'us-west-2', },
            'sae1': {'sa-east-1', },
            'apn1': {'ap-northeast-1', },
            'apn2': {'ap-northeast-2', },
            'aps1': {'ap-southeast-1', },
            'ap22': {'ap-southeast-2', },
            'euw1': {'eu-west-1', },
            'euc1': {'eu-central-1', },
        }

        region_name = region_name.lower() if region_name else cls.get_region()

        for shortname, candidates in alternatives.items():
            if region_name in candidates:
                return shortname

        return region_name



# -*- coding: utf-8 -*-

import tempfile
import shutil

from os.path import dirname

def render(template, mapping):
    """
    Render template document aginst mapping dictionary with jinja2.

    :param str template:
    :param dict mapping:
    :return:
    :rtype: str
    """
    from jinja2 import Environment, DictLoader

    env = Environment(loader=DictLoader({
        'cfn': str(template)
    }))
    tpl = env.get_template('cfn')

    return tpl.render(mapping)


class tempdir(object):
    """
    Context manager to create and handle temp dir, expose utility methods,
    and cleanup on exit.
    """
    def __init__(self, base_path=None, cleanup=True):
        self.__base_path = base_path
        self.__cleanup = cleanup

    @property
    def base_path(self):
        """
        Base path for temp files

        :return:
        """
        return self.__base_path

    @property
    def path(self):
        return self.__path

    def cleanup(self):
        try:
            shutil.rmtree(self.path)
        except AttributeError:
            pass  # Never `enter`ed, no path to delete

    def __enter__(self):
        self.__path = tempfile.mkdtemp(dir=self.__base_path)
        if self.__base_path is None:
            self.__base_path = dirname(self.__path)
        return self

    def __exit__(self, *_):
        if self.__cleanup:
            self.cleanup()



from __future__ import print_function

import boto3
import botocore.exceptions

from aws import Aws

class S3(Aws):
    def __init__(self, s3_resource=None):
        self._s3r = s3_resource or boto3.resource('s3')

    def safe_create_bucket(self, bucket_name, region=None):
        """
        Create S3 bucket in the specified region if it's not there,
        and return the boto3 bucket object.
        """
        region = region or self.get_region()

        args = {"Bucket": bucket_name}
        if region not in (None, 'us-east-1'):
            args.update({
                "CreateBucketConfiguration": {"LocationConstraint": region}
            })

        try:
            return self._s3r.create_bucket(**args)
        except botocore.exceptions.ClientError:
            pass

        return self._s3r.Bucket(bucket_name)

    def upload(self, source, bucket_name, s3_key):
        """
        :param str source: file path or body
        :param bucket_name:
        :param s3_key:
        :return:
        """
        obj = self.safe_create_bucket(bucket_name).Object(s3_key)

        with open(source, 'rb') as fp:
            obj.put(Body=fp)

        return s3_key

    def cleanup_bucket(self, bucket_name):
        print("  * Cleaning up bucket `%s`" % bucket_name)
        bucket = self._s3r.Bucket(bucket_name)

        bucket.delete_objects(
            Delete={
                "Objects": [
                    {"Key": obj.key} for obj in bucket.objects.all()
                ]
            }
        )
